"""
Segmented regression
--------------------

Broken line regression with two segments.
"""

import numpy as np

__all__ = [
    "segreg",
    "predict",
]


def _ols(Xi, Y):
    XT_X_inv = np.linalg.inv(Xi.T @ Xi)
    params = XT_X_inv @ (Xi.T @ Y)
    return params


def segreg(x, Y, psi0, tol=1e-5, maxiter=30):
    r"""Segmented regression with one breakpoint.

    Parameters
    ----------
    x, Y : (M,) ndarray
        Data points.
    psi0 : scalar
        Initial guess for breakpoint coordinate.
    tol : float, default=1e-5
        Convergence tolerance.
    maxiter : int, default=30
        Force break after this iterations.

    Returns
    -------
    params : (4,) ndarray
        Estimated parameters: b0, b1, b2, psi.
    reached_max : bool
        Iteration is finished not by convergence but by reaching maximum iteration.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.segreg import segreg, predict
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     Y = next(data.profiles())
    ...     X = data.x()[:len(Y)]
    >>> x = X[X < 10]
    >>> (b0, b1, b2, psi1), _ = segreg(x, Y[:len(x)], 7)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(X, Y.T, color="gray")
    ... plt.plot(x, predict(x, b0, b1, b2, psi1))

    """
    Xi = np.array(
        [
            np.ones_like(x),
            x,
            (x - psi0) * np.heaviside(x - psi0, 0),
            -np.heaviside(x - psi0, 0),
        ]
    ).T

    b0, b1, b2, gamma = _ols(Xi, Y)
    RSS = np.sum((Y - predict(x, b0, b1, b2, psi0)) ** 2)

    psi_converged = False
    for _ in range(maxiter):
        RSS_new = RSS
        lamda = 1
        while True:
            psi0_new = psi0 + lamda * gamma / b2
            RSS_new = np.sum((Y - predict(x, b0, b1, b2, psi0_new)) ** 2)
            lamda /= 2

            if (psi0_new <= x[0]) or (psi0_new >= x[-1]):
                # exceeded domain; make step size smaller
                continue
            if RSS_new >= RSS:
                # RSS not decreased; make step size smaller
                continue
            psi_converged = np.abs(psi0 - psi0_new) <= tol
            if psi_converged:
                break

        if not psi_converged:
            psi_converged = np.abs(psi0 - psi0_new) <= tol
        if psi_converged:
            psi0 = psi0_new
            reached_max = False
            break

        psi0 = psi0_new
        RSS = RSS_new
        Xi[:, 2] = (x - psi0) * np.heaviside(x - psi0, 0)
        Xi[:, 3] = -np.heaviside(x - psi0, 0)
        b0, b1, b2, gamma = _ols(Xi, Y)
    else:
        reached_max = True

    params = np.array([b0, b1, b2, psi0_new])
    return params, reached_max


def predict(x, b0, b1, b2, psi):
    r"""Predict y values from x coordinates by segmented regression model.

    The model is

    .. math::

        \beta_0 + \beta_1 + \beta_2 (\xi - \psi)_+.

    Parameters
    ----------
    x : (M,) ndarray
        X coordinates.
    b0, b1, b2, psi : scalar
        Model parameters.

    Returns
    -------
    (M,) ndarray
        Predicted y coordinates.
    """
    x = np.asarray(x)
    return b0 + b1 * x + b2 * (x - psi) * np.heaviside(x - psi, 0)



"""Commands to process profiles."""

import pathlib
from importlib.metadata import entry_points

from heavyedge.cli.command import Command, register_command

PLUGIN_ORDER = 0.0


@register_command("prep", "Preprocess raw profiles")
class PrepCommand(Command):
    def add_parser(self, main_parser):
        prep = main_parser.add_parser(
            self.name,
            description="Preprocess raw profiles and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        prep.add_argument(
            "--type",
            choices=[ep.name for ep in entry_points(group="heavyedge.rawdata")],
            required=True,
            help="Type of the raw profile data.",
        )
        prep.add_argument("--name", help="Name to label output dataset.")
        prep.add_argument(
            "raw",
            type=pathlib.Path,
            help="Path to raw profile data.",
        )
        prep.add_config_argument(
            "--res",
            type=float,
            help=(
                "Spatial resolution of profile data, i.e., points per unit length. "
                "Length unit must match the profile height unit."
            ),
        )
        prep.add_config_argument(
            "--sigma",
            type=float,
            help="Standard deviation of Gaussian kernel.",
        )
        prep.add_config_argument(
            "--std-thres",
            type=float,
            help="Standard deviation threshold for contact point detection.",
        )
        prep.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        import numpy as np

        from heavyedge.api import preprocess
        from heavyedge.io import ProfileData

        self.logger.info(f"Preprocessing: {args.raw}")

        raw_type = entry_points(group="heavyedge.rawdata")[args.type].load()
        raw = raw_type(args.raw)
        raw_profiles = zip(raw.profiles(), raw.profile_names())

        # search for the first valid profile
        for profile, name in raw_profiles:
            if len(profile) == 0:
                continue
            if np.any(np.isnan(profile)):
                continue
            M = len(profile)
            break

        with ProfileData(args.output, "w").create(M, args.res, args.name) as out:
            # write the first valid profile
            Y, L = preprocess(profile, args.sigma, args.std_thres)
            out.write_profiles(Y.reshape(1, -1), [L], [name])

            for profile, name in raw_profiles:
                if len(profile) == 0:
                    continue
                if np.any(np.isnan(profile)):
                    continue
                Y, L = preprocess(profile, args.sigma, args.std_thres)
                out.write_profiles(Y.reshape(1, -1), [L], [name])

        self.logger.info(f"Preprocessed: {out.path}")


@register_command("outlier", "Filter outlier profiles")
class OutlierCommand(Command):
    def add_parser(self, main_parser):
        outlier = main_parser.add_parser(
            self.name,
            description="Remove outlier profiles and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        outlier.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        outlier.add_config_argument(
            "--z",
            type=float,
            help="Modified Z-score threshold for outlier detection.",
        )
        outlier.add_argument(
            "-o", "--output", type=pathlib.Path, help="Output file path"
        )

    def run(self, args):
        from heavyedge import ProfileData
        from heavyedge.api import outlier

        self.logger.info(f"Removing outliers: {args.profiles}")

        with ProfileData(args.profiles) as data:
            _, M = data.shape()
            res = data.resolution()
            name = data.name()
            is_outlier = outlier(data.profiles(), args.z)

            with ProfileData(args.output, "w").create(M, res, name) as out:
                for skip, Y, L, name in zip(
                    is_outlier,
                    data._file["profiles"],
                    data._file["len"],
                    data.profile_names(),
                ):
                    if not skip:
                        out.write_profiles(Y.reshape(1, -1), [L], [name])

        self.logger.info(f"Removed outliers: {out.path}")


@register_command("mean", "Compute mean profile")
class MeanCommand(Command):
    def add_parser(self, main_parser):
        mean = main_parser.add_parser(
            self.name,
            description="Compute mean profile and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        mean.add_argument(
            "profiles",
            type=pathlib.Path,
            help="Path to preprocessed profile data in 'ProfileData' structure.",
        )
        mean.add_config_argument(
            "--wnum",
            type=int,
            help="Number of sample points to compute Wasserstein mean.",
        )
        mean.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        import numpy as np

        from heavyedge.api import mean
        from heavyedge.io import ProfileData

        self.logger.info(f"Averaging: {args.profiles}")

        with ProfileData(args.profiles) as data:
            _, M = data.shape()
            x = data.x()
            res = data.resolution()
            name = data.name()
            pmean = mean(x, data.profiles(), args.wnum)

        with ProfileData(args.output, "w").create(M, res, name) as out:
            L = len(pmean)
            pmean = np.pad(pmean, (0, M - L), constant_values=np.nan)
            out.write_profiles(pmean.reshape(1, -1), [L], [name])

        self.logger.info(f"Averaged: {out.path}")


@register_command("merge", "Merge profile data")
class MergeCommand(Command):
    def add_parser(self, main_parser):
        merge = main_parser.add_parser(
            self.name,
            description="Merge profile data and save as hdf5 file.",
            epilog="The resulting hdf5 file is in 'ProfileData' structure.",
        )
        merge.add_argument(
            "profiles",
            nargs="+",
            type=pathlib.Path,
            help="Paths to preprocessed profile data in 'ProfileData' structure.",
        )
        merge.add_argument("--name", help="Name to label output dataset.")
        merge.add_argument("-o", "--output", type=pathlib.Path, help="Output file path")

    def run(self, args):
        from heavyedge.io import ProfileData

        with ProfileData(args.profiles[0]) as data:
            _, M = data.shape()
            res = data.resolution()

        with ProfileData(args.output, "w").create(M, res, args.name) as out:
            for p in args.profiles:
                with ProfileData(p) as data:
                    out.write_profiles(
                        data._file["profiles"][:],
                        data._file["len"][:],
                        data._file["names"][:],
                    )



import os
import subprocess


def test_process_commands(tmp_rawdata_type2_path, tmp_path):
    processed_path = tmp_path / "ProcessedProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "prep",
            "--type",
            "csvs",
            "--res=1",
            "--sigma=32",
            "--std-thres=0.1",
            tmp_rawdata_type2_path,
            "-o",
            processed_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(processed_path)

    filtered_path = tmp_path / "FilteredProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "outlier",
            "--z",
            "3.5",
            processed_path,
            "-o",
            filtered_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(filtered_path)

    mean_path = tmp_path / "MeanProfile.h5"
    subprocess.run(
        [
            "heavyedge",
            "mean",
            "--wnum",
            "100",
            processed_path,
            "-o",
            mean_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(mean_path)

    merged_path = tmp_path / "MergedProfiles.h5"
    subprocess.run(
        [
            "heavyedge",
            "merge",
            processed_path,
            processed_path,
            "-o",
            merged_path,
        ],
        capture_output=True,
        check=True,
    )
    assert os.path.exists(merged_path)



"""Landmark detection."""

import numpy as np
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks, peak_prominences

from ..segreg import segreg

__all__ = [
    "landmarks_type2",
    "landmarks_type3",
    "plateau_type2",
    "plateau_type3",
]


def landmarks_type2(Y, sigma):
    """Find landmarks for heavy edge profile without trough.

    Parameters
    ----------
    Y : 1-D array
        1-dimensional heavy edge profile data.
        The last point must be the contact point.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.

    Returns
    -------
    landmarks : (3,) array of int
        Indices of landmarks.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import landmarks_type2
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     Y = next(data.profiles())
    >>> lm = landmarks_type2(Y, 32)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Y)
    ... plt.plot(lm, Y[lm], "o")
    """
    cp = len(Y) - 1

    Y_smooth = gaussian_filter1d(Y, sigma)
    peaks, _ = find_peaks(Y_smooth)
    peak = peaks[-1]

    Y_ = Y_smooth[:peak]
    pts = np.column_stack([np.arange(len(Y_)), Y_])
    x, y = pts - pts[0], pts[-1] - pts[0]
    dists = x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]
    slope = np.diff(dists)
    (extrema,) = np.nonzero(np.diff(np.sign(slope)))
    K_pos = extrema[slope[extrema] > 0]
    knee = K_pos[np.argmax(np.abs(dists[K_pos]))]

    return np.array([cp, peak, knee])


def landmarks_type3(Y, sigma):
    """Find landmarks for heavy edge profile with trough.

    Parameters
    ----------
    Y : 1-D array
        1-dimensional heavy edge profile data.
        The last point must be the contact point.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.

    Returns
    -------
    landmarks : (4,) array of int
        Indices of landmarks.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import landmarks_type3
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as data:
    ...     Y = next(data.profiles())
    >>> lm = landmarks_type3(Y, 32)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Y)
    ... plt.plot(lm, Y[lm], "o")
    """
    cp = len(Y) - 1

    Y_smooth = gaussian_filter1d(Y, sigma)
    peaks, _ = find_peaks(Y_smooth)
    peak = peaks[-1]

    troughs, _ = find_peaks(-Y_smooth)
    troughs = troughs[troughs < peak]

    if len(troughs) > 0:
        prominences = peak_prominences(-Y_smooth, troughs)[0]
        most_prominent_idx = np.argmax(prominences)
        trough = troughs[most_prominent_idx]

        Y_ = Y_smooth[: int(trough) + 1]
        pts = np.column_stack([np.arange(len(Y_)), Y_])
        x, y = pts - pts[0], pts[-1] - pts[0]
        dists = x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]
        slope = np.diff(dists)
        (extrema,) = np.nonzero(np.diff(np.sign(slope)))
        K_neg = extrema[slope[extrema] < 0]
        knee = K_neg[np.argmax(np.abs(dists[K_neg]))]

    else:
        Y_ = Y_smooth[:peak]
        pts = np.column_stack([np.arange(len(Y_)), Y_])
        x, y = pts - pts[0], pts[-1] - pts[0]
        dists = x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]
        slope = np.diff(dists)
        (extrema,) = np.nonzero(np.diff(np.sign(slope)))
        K_pos = extrema[slope[extrema] > 0]
        knee, trough = K_pos[np.argmax(np.abs(dists[K_pos]))]

    return np.array([cp, peak, trough, knee])


def plateau_type2(x, Y, peak, knee):
    """Find plateau for heavy edge profile without trough.

    Parameters
    ----------
    x : (M,) array
        Spatial coordinates of profile data.
    Y : (M,) array
        1-dimensional heavy edge profile data.
        The last point must be the contact point.
    peak, knee : int
        Peak and knee point indices.

    Returns
    -------
    b0, b1, psi : scalar
        Plateau height, slope and boundary.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import plateau_type2
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     Y = next(data.profiles())
    ...     x = data.x()[:len(Y)]
    >>> b0, b1, psi = plateau_type2(x, Y, 2000, 1300)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Y, color="gray", alpha=0.2)
    ... X = x[x < psi]
    ... plt.plot(X, b0 + b1 * X)
    """
    (b0, b1, _, psi), _ = segreg(x[:peak], Y[:peak], x[knee])
    if b1 < 0:
        psi_idx = knee + np.argmin(np.abs(Y[knee:peak] - b0))
        b1 = 0.0
        psi = x[psi_idx]
    return (b0, b1, psi)


def plateau_type3(x, Y, trough, knee):
    """Find plateau for heavy edge profile with trough.

    Parameters
    ----------
    x : (M,) array
        Spatial coordinates of profile data.
    Y : (M,) array
        1-dimensional heavy edge profile data.
        The last point must be the contact point.
    peak, knee : int
        Trough and knee point indices.

    Returns
    -------
    b0, b1, psi : scalar
        Plateau height, slope and boundary.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import plateau_type3
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as data:
    ...     Y = next(data.profiles())
    ...     x = data.x()[:len(Y)]
    >>> b0, b1, psi = plateau_type3(x, Y, 1500, 1000)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Y, color="gray", alpha=0.2)
    ... X = x[x < psi]
    ... plt.plot(X, b0 + b1 * X)
    """
    (b0, b1, _, psi), _ = segreg(x[:trough], Y[:trough], x[knee])
    if b1 > 0:
        psi_idx = knee + np.argmin(np.abs(Y[knee:trough] - b0))
        b1 = 0.0
        psi = x[psi_idx]
    return (b0, b1, psi)



"""
Wasserstein distance
--------------------

Wasserstein-related functions.
"""

import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import interp1d

from ._wasserstein import optimize_q

__all__ = [
    "quantile",
    "wdist",
    "wmean",
]


def quantile(x, f, t):
    """Convert a probability distribution to a quantile function.

    Parameters
    ----------
    x : ndarray
        Points over which *f* is measured.
    f : ndarray
        The empirical probability density function.
    t : ndarray
        Points over which the quantile function will be measured.
        Must be strictly increasing from 0 to 1.

    Returns
    -------
    ndarray
        Quantile function of *f* over *t*.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.wasserstein import quantile
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     Y = next(data.profiles())
    ...     x = data.x()[:len(Y)]
    >>> f = Y / np.trapezoid(Y, x)
    >>> t = np.linspace(0, 1, 100)
    >>> Q = quantile(x[:len(f)], f, t)
    """
    G = cumulative_trapezoid(f, x, initial=0)
    return interp1d(G, x, bounds_error=False, fill_value=(x[0], x[-1]))(t)


def wdist(x1, f1, x2, f2, grid_num):
    r"""Wasserstein distance between two 1D probability distributions.

    .. math::

        d_W(f_1, f_2)^2 = \int^1_0 (Q_1(t) - Q_2(t))^2 dt

    where :math:`Q_i` is the quantile function of :math:`f_i`.

    Parameters
    ----------
    x1, f1 : ndarray
        The first empirical probability density function.
    x2, f2 : ndarray
        The second empirical probability density function.
    grid_num : int
        Number of sample points in [0, 1] to approximate the integral.

    Returns
    -------
    scalar
        Wasserstein distance.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.wasserstein import wdist
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     (Y1, Y2), (L1, L2), _ = data[:2]
    >>> x1, Y1 = x[:L1], Y1[:L1]
    >>> x2, Y2 = x[:L2], Y2[:L2]
    >>> f1, f2 = Y1 / np.trapezoid(Y1, x1), Y2 / np.trapezoid(Y2, x2)
    >>> d = wdist(x1, f1, x2, f2, 100)
    """
    grid = np.linspace(0, 1, grid_num)
    Q1 = quantile(x1, f1, grid)
    Q2 = quantile(x2, f2, grid)
    return np.trapezoid((Q1 - Q2) ** 2, grid) ** 0.5


def wmean(xs, fs, grid_num):
    """Fréchet mean of probability distrubutions using Wasserstein metric.

    Parameters
    ----------
    xs : list of ndarray
        Points over which each distribution in *fs* is measured.
    fs : list of ndarray
        Empirical probability density functions.
    grid_num : int
        Number of sample points in [0, 1] to approximate the integral.

    Returns
    -------
    x, f : ndarray
        Probability density function.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.wasserstein import wmean
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     (Y1, Y2), (L1, L2), _ = data[:2]
    >>> x1, Y1 = x[:L1], Y1[:L1]
    >>> x2, Y2 = x[:L2] + 3, Y2[:L2]
    >>> f1, f2 = Y1 / np.trapezoid(Y1, x1), Y2 / np.trapezoid(Y2, x2)
    >>> x, f = wmean([x1, x2], [f1, f2], 100)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x1, f1, "--", color="gray")
    ... plt.plot(x2, f2, "--", color="gray")
    ... plt.plot(x, f)
    """
    grid = np.linspace(0, 1, grid_num)
    Q = np.array([quantile(x, f, grid) for x, f in zip(xs, fs)])
    g = np.mean(Q, axis=0)
    if np.all(np.diff(g) >= 0):
        q = g
    else:
        q = optimize_q(g)
    pdf = 1 / np.gradient(q, grid)
    pdf[-1] = 0
    pdf /= np.trapezoid(pdf, q)
    return q, pdf



"""Helper functions for wasserstein distance."""

cimport cython
cimport numpy as cnp
from libc.stdlib cimport free, malloc

import numpy as np

cnp.import_array()

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef cnp.ndarray[cnp.float64_t, ndim=1] optimize_q(double[:] g):
    # M = number of probability-space grid, N = number of distance-space grid
    cdef Py_ssize_t i, j, idx
    cdef double[:] y_vals = np.unique(g)
    cdef Py_ssize_t M = g.shape[0], N = y_vals.shape[0]
    # Should memory error occurs, may need to make ca 1d and overwrite during loop.
    cdef double *ca = <double *> malloc(M * N * sizeof(double))
    if not ca:
        raise MemoryError()
    cdef int *predecessor = <int *> malloc((M - 1) * N * sizeof(int))
    if not predecessor:
        raise MemoryError()

    # Compute costs
    for i in range(M):  # TODO: parallize this i-loop
        for j in range(N):
            ca[i * N + j] = (g[i] - y_vals[j]) ** 2

    # Accumulate costs
    cdef Py_ssize_t prev_min_j
    cdef double prev_min
    for i in range(1, M):
        prev_min_j = 0
        prev_min = ca[(i - 1) * N + prev_min_j]
        for j in range(N):
            if ca[(i - 1) * N + j] < prev_min:
                prev_min_j = j
                prev_min = ca[(i - 1) * N + j]
            ca[i * N + j] += prev_min
            predecessor[(i - 1) * N + j] = prev_min_j

    # Backtrack
    cdef cnp.ndarray[cnp.float64_t, ndim=1] q = np.empty(M, dtype=np.float64)
    idx = 0
    # Last column
    for j in range(1, N):
        if ca[(M - 1) * N + j] < ca[(M - 1) * N + idx]:
            idx = j
    q[M - 1] = y_vals[idx]
    free(ca)
    # Other columns
    for i in range(1, M):
        idx = predecessor[(M - 1 - i) * N + idx]
        q[M - 1 - i] = y_vals[idx]

    free(predecessor)
    return q



"""Raw profile data files."""

import abc
import csv
import numbers
from collections.abc import Sequence
from pathlib import Path

import numpy as np

__all__ = [
    "RawProfileBase",
    "RawProfileCsvs",
]


class RawProfileBase(abc.ABC):
    """Base class to read raw profile data.

    All raw profiles must have the same length.

    Notes
    -----
    ``self[key]`` returns a tuple of profile(s) and profile name(s).
    """

    def __init__(self, path):
        self.path = Path(path).expanduser()

    def __len__(self):
        return self.count_profiles()

    def __getitem__(self, key):
        """Profile and name indexing.

        By default, uses :meth:`all_profiles` and :meth:`profile_names`.
        Subclasses are encouraged to redefine this method for better performance.
        """
        return (self.all_profiles()[key], np.array(self.profile_names())[key])

    @abc.abstractmethod
    def count_profiles(self):
        """Number of raw profiles.

        Returns
        -------
        int
        """

    @abc.abstractmethod
    def profiles(self):
        """Yield raw profiles.

        Yields
        ------
        1-D ndarray
        """

    def all_profiles(self):
        """Return all profiles as an 2-D array.

        Returns
        -------
        2-D ndarray
        """
        return np.array([p for p in self.profiles()])

    @abc.abstractmethod
    def profile_names(self):
        """Yield profile names.

        Yields
        ------
        str
        """


class RawProfileCsvs(RawProfileBase):
    """Read raw profile data from a directory containing CSV files.

    Directory structure:

    .. code-block::

        rawdata/
        ├── profile1.csv
        ├── profile2.csv
        └── ...

    Parameters
    ----------
    path : pathlike
        Path to the directory containing the raw CSV files.

    Notes
    -----
    - Each CSV file must contain a single column of numeric values (no header).
    - The order of profiles is determined by the sorted filenames.
    - The profile name is derived from the filename stem.

    Examples
    --------
    >>> from heavyedge import get_sample_path, RawProfileCsvs
    >>> profiles = RawProfileCsvs(get_sample_path("Type3")).all_profiles()
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... for Y in profiles:
    ...     plt.plot(Y)
    """

    def __init__(self, path):
        super().__init__(path)
        self._files = sorted(self.path.glob("*.csv"))

    @staticmethod
    def _read_profile(path):
        with open(path, newline="") as csvfile:
            reader = csv.reader(csvfile)
            profile = np.array([float(row[0]) for row in reader])
        return profile

    def __getitem__(self, key):
        if isinstance(key, numbers.Integral):
            file = self._files[key]
            return (self._read_profile(file), str(file.stem))
        elif isinstance(key, slice):
            files = self._files[key]
            profiles, names = [], []
            for file in files:
                profiles.append(self._read_profile(file))
                names.append(str(file.stem))
            return (np.array(profiles), np.array(names))
        elif isinstance(key, (Sequence, np.ndarray)):
            profiles, names = [], []
            for k in key:
                file = self._files[k]
                profiles.append(self._read_profile(file))
                names.append(str(file.stem))
            return (np.array(profiles), np.array(names))
        else:
            raise TypeError(f"Invalid index type: {type(key)}")

    def count_profiles(self):
        return len(list(self._files))

    def profiles(self):
        for file in self._files:
            yield self._read_profile(file)

    def profile_names(self):
        for f in self._files:
            yield str(f.stem)



"""Argument parser."""

import argparse

import yaml

__all__ = [
    "ConfigFileAction",
    "ConfigArgumentParser",
]


class ConfigFileAction(argparse.Action):
    """Action for an argument to load values from YAML config file.

    The config file consists of key-value pairs for arguments.
    Explicitly passed argument overloads the config file value.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(parser, ConfigArgumentParser):
            config_args = [arg.dest for arg in parser.config_arguments()]
        else:
            config_args = []
        if values is None:
            return
        with values as f:
            data = {
                key.replace("-", "_"): val for (key, val) in yaml.full_load(f).items()
            }
        for k, v in vars(namespace).items():
            if v is None and k in config_args and k in data:
                setattr(namespace, k, data[k])


class ConfigArgumentParser(argparse.ArgumentParser):
    """Argument parser which can read config file.

    Use :meth:`add_config_argument` to add an argument which can be read from the
    config file. When the first config argument is added, the ``--config`` option is
    automatically added.

    Examples
    --------
    >>> from heavyedge.cli import ConfigArgumentParser
    >>> parser = ConfigArgumentParser("foo")
    >>> _ = parser.add_config_argument("--bar")
    >>> parser.print_help()
    usage: foo [-h] [--config CONFIG] [--bar BAR]
    <BLANKLINE>
    options:
    -h, --help       show this help message and exit
    --config CONFIG  YAML file specifying config options.
    <BLANKLINE>
    config options:
    --bar BAR
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._config_argument_group = self.add_argument_group("config options")
        self._config_argument = []

    def config_arguments(self):
        return self._config_argument

    def add_config_argument(self, name, **kwargs):
        """Add argument which can be read from the config file.

        Parameters
        ----------
        name : str
            Name of the argument.
        kwargs : dict
            Additional arguments passed to :meth:`argparse.ArgumentParser.add_argument`.
        """
        kwargs.update(required=False, default=None)
        if len(self.config_arguments()) == 0:
            # First config argument: add config parser first
            self.add_argument(
                "--config",
                type=open,
                action=ConfigFileAction,
                help="YAML file specifying config options.",
            )
        ret = self._config_argument_group.add_argument(name, **kwargs)
        self._config_argument.append(ret)
        return ret



"""High-level Python runtime interface."""

__all__ = [
    "preprocess",
    "outlier",
    "mean",
    "landmarks_type2",
    "landmarks_type3",
    "plateau_type2",
    "plateau_type3",
]

from .landmarks import landmarks_type2, landmarks_type3, plateau_type2, plateau_type3
from .profile import mean, outlier, preprocess



import numpy as np

from heavyedge import ProfileData, RawProfileCsvs


def test_RawData_dunder(tmp_rawdata_type2_path):
    data = RawProfileCsvs(tmp_rawdata_type2_path)
    assert len(data) == data.count_profiles()

    item = data[np.int64(0)]
    profile = next(data.profiles())
    name = next(data.profile_names())
    assert np.all(item[0] == profile)
    assert item[1] == name

    item = data[:1]
    assert np.all(item[0] == [profile])
    assert np.all(item[1] == [name])

    item = data[np.array([0])]
    assert np.all(item[0] == [profile])
    assert np.all(item[1] == [name])

    item = data[:0]
    assert item[0].size == 0
    assert len(item[1]) == 0


def test_ProfileData_dunder(tmp_prepdata_type2_path):
    with ProfileData(tmp_prepdata_type2_path) as data:
        assert len(data) == data.shape()[0]

        item = data[np.int64(0)]
        profile = data.all_profiles()[0]
        length = len(next(data.profiles()))
        name = next(data.profile_names())
        assert np.all(item[0] == profile)
        assert item[1] == length
        assert item[2] == name

        item = data[:1]
        assert np.all(item[0] == [profile])
        assert np.all(item[1] == [length])
        assert np.all(item[2] == [name])

        item = data[np.array([0])]
        assert np.all(item[0] == [profile])
        assert np.all(item[1] == [length])
        assert np.all(item[2] == [name])

        item = data[:0]
        assert item[0].size == 0
        assert item[1].size == 0
        assert item[2].size == 0



"""Central package for heavy edge analysis."""

__all__ = [
    "get_sample_path",
    "RawProfileCsvs",
    "ProfileData",
]

import logging

from .io import (
    ProfileData,
    RawProfileCsvs,
)
from .samples import get_sample_path

logger = logging.getLogger(__name__)


def main():
    import argparse
    import sys
    from importlib.metadata import entry_points

    from heavyedge.cli.command import REGISTERED_COMMANDS

    from .cli.parser import ConfigArgumentParser

    def format_epilog(commands):
        command_lengths = []
        for _, command_dict in commands:
            for name in command_dict.keys():
                command_lengths.append(len(name))

        INDENT1 = 2 * " "
        INDENT2 = 4 * " "
        SPACING = max(max(command_lengths) + 5, 20)
        ret = "LIST OF COMMANDS\n"
        for group, command_dict in commands:
            ret += f"{INDENT1}{group}:\n"
            for name, (_, desc) in command_dict.items():
                ret += f"{INDENT2}{name}{(SPACING - len(name)) * ' '}{desc}\n"
            ret += "\n"
        return ret

    ORDERS = []
    COMMANDS = []
    for ep in entry_points(group="heavyedge.commands"):
        module = ep.load()
        order = getattr(module, "PLUGIN_ORDER", None)
        commands = dict()
        for key in list(REGISTERED_COMMANDS.keys()):
            commands[key] = REGISTERED_COMMANDS.pop(key)
        ORDERS.append(order)
        COMMANDS.append((ep.name, commands))

    max_order = max([x for x in ORDERS if x is not None])
    ORDERS = [x if x is not None else max_order for x in ORDERS]
    COMMANDS = [x[1] for x in sorted(zip(ORDERS, COMMANDS))]

    heavyedge_parser = argparse.ArgumentParser(
        prog="heavyedge",
        description="Heavy edge profile analysis.",
        epilog=format_epilog(COMMANDS),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    heavyedge_parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        help="Set logging level.",
    )
    subparser = heavyedge_parser.add_subparsers(
        title="Command",
        dest="command",
        metavar="<command>",
        help="Refer to the epilog for a list of commands.",
        parser_class=ConfigArgumentParser,
    )

    COMMAND_OBJS = dict()
    for _, command_dict in COMMANDS:
        for name, (cls, _) in command_dict.items():
            COMMAND_OBJS[name] = cls(name, logger)

    for command in COMMAND_OBJS.values():
        command.add_parser(subparser)

    args = heavyedge_parser.parse_args()

    logging.basicConfig(
        format="heavyedge: [%(asctime)s] [%(levelname)8s] --- %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=args.log_level.upper(),
    )
    logger.debug(f"Input command: {' '.join(sys.argv)}")

    if args.command is None:
        heavyedge_parser.print_help(sys.stderr)
        sys.exit(1)

    else:
        command = COMMAND_OBJS[args.command]
        try:
            command.run(args)
        except Exception:
            input_command = " ".join(sys.argv)
            logger.exception(f"Got exception while running '{input_command}'.")
            sys.exit(1)



"""Processed profile data files."""

import numbers
from collections.abc import Sequence
from pathlib import Path

import h5py
import numpy as np

__all__ = [
    "ProfileData",
]


class ProfileData:
    """Preprocessed 1-dimensional profile data as hdf5 file.

    Parameters
    ----------
    path : pathlike
        Path to the hdf5 file.
    mode : {'r', 'w', 'r+', 'a', 'w-'}
        Mode to open the file.
    kwargs : dict
        Optional arguments passed to :class:`h5py.File`.

    Notes
    -----
    ``self[key]`` returns a tuple of full profile data, profile length(s) and
    profile name(s).
    """

    def __init__(self, path, mode="r", **kwargs):
        self.path = Path(path).expanduser()
        self._file = h5py.File(path, mode, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, type, value, trace_back):
        self._file.close()

    def __len__(self):
        return self.shape()[0]

    def __getitem__(self, key):
        if isinstance(key, numbers.Integral):
            profile = self._file["profiles"][key]
            length = self._file["len"][key]
            name = str(self._file["names"][key], encoding="utf-8")
            return (profile, length, name)
        elif isinstance(key, (slice, Sequence, np.ndarray)):
            profiles = self._file["profiles"][key]
            lengths = self._file["len"][key]
            names = np.char.decode(
                self._file["names"][key].astype("S"), encoding="utf-8"
            )
            return (profiles, lengths, names)
        else:
            raise TypeError(f"Invalid index type: {type(key)}")

    def close(self):
        self._file.close()

    def create(self, M, resolution, name=None):
        """Create datasets and write metadata.

        Parameters
        ----------
        M : int
            Maximum length of profile data.
        resolution : float
            Spatial resolution of the profile data.
        name : str, optional
            Unique name to identify the dataset.

        Returns
        -------
        obj
            Returns the object itself.
        """
        if name is None:
            name = str(self.path.with_suffix(""))
        self._file.attrs["name"] = name
        self._file.attrs["res"] = resolution

        self._file.create_dataset(
            "profiles",
            (0, M),
            maxshape=(None, M),
            dtype=float,
        )
        self._file.create_dataset(
            "len",
            (0,),
            maxshape=(None,),
            dtype=int,
        )
        self._file.create_dataset(
            "names",
            (0,),
            maxshape=(None,),
            dtype=h5py.string_dtype(),
        )
        return self

    def name(self):
        """Unique name of the dataset.

        Returns
        -------
        str
        """
        return self._file.attrs["name"]

    def resolution(self):
        """Spatial resolution of the profile data.

        Returns
        -------
        float
        """
        return self._file.attrs["res"]

    def shape(self):
        """Shape of profile dataset.

        Returns
        -------
        (N, M)
        """
        return self._file["profiles"].shape

    def x(self):
        """Spatial coordinates.

        Returns
        -------
        (M,) ndarray
        """
        return np.arange(self.shape()[1]) / self.resolution()

    def write_profiles(self, profiles, lengths, names):
        """Append profiles data to file.

        Parameters
        ----------
        profiles : (N, M) ndarray of float
            1-dimensional profile.
        lengths : (N,) array of int
            Number of data in *profiles* from reference point to contact point.
        names : list of str
            Profile names.
        """
        N = len(profiles)

        dset = self._file["profiles"]
        dset.resize(dset.shape[0] + N, axis=0)
        dset[-N:] = profiles

        dset = self._file["len"]
        dset.resize(dset.shape[0] + N, axis=0)
        dset[-N:] = lengths

        dset = self._file["names"]
        dset.resize(dset.shape[0] + N, axis=0)
        dset[-N:] = names

    def profiles(self):
        """Yield profiles.

        Profiles are cropped by the contact point.

        Yields
        ------
        1-D ndarray
        """
        for profile, length in zip(self._file["profiles"], self._file["len"]):
            yield profile[:length]

    def profile_names(self):
        """Yield profile names.

        Yields
        ------
        str
        """
        for name in self._file["names"]:
            yield str(name, encoding="utf-8")

    def all_profiles(self):
        """Return all profiles.

        Returns
        -------
        (N, M) ndarray
            All N profiles data.
        """
        return self._file["profiles"][:]



"""Function to register commands."""

import abc

__all__ = [
    "REGISTERED_COMMANDS",
    "Command",
    "register_command",
]


REGISTERED_COMMANDS = dict()


class Command(abc.ABC):
    """Sub-command for CLI interface.

    Parameters
    ----------
    name : str
        Name of the sub-command.
    logger : logging.Logger
        Logger to log the command run.

    Examples
    --------
    >>> class MyCommand(Command):
    ...     def add_parser(self, main_parser):
    ...         parser = main_parser.add_parser(self.name)
    ...         parser.add_argument("foo")
    ...     def run(self, args):
    ...         self.logger.info("Run my command")
    ...         print(args.foo)
    """

    def __init__(self, name, logger):
        self._name = name
        self._logger = logger

    @property
    def name(self):
        """Name of the command for :meth:`add_parser`."""
        return self._name

    @property
    def logger(self):
        """Logger for :meth:`run`."""
        return self._logger

    @abc.abstractmethod
    def add_parser(self, main_parser):
        """Add the command parser to main parser.

        Parameters
        ----------
        main_parser : argparse._SubParsersAction
            Subparser constructor having :class:`heavyedge.cli.ConfigArgumentParser` as
            parser class.
        """
        ...

    @abc.abstractmethod
    def run(self, args):
        """Run the command.

        Parameters
        ----------
        args : argparse.Namespace
        """
        ...


def register_command(name, desc):
    """Decorator to register the command class for the argument parser.

    Parameters
    ----------
    name : str
        The unique name of the command.
    desc : str
        A short description of the command's purpose.

    Examples
    --------
    Decorate the class definition.

    >>> from heavyedge.cli import Command, register_command
    >>> @register_command("foo", "My command")
    ... class MyCommand(Command):
    ...     ...

    See Also
    --------
    heavyedge.cli.Command
    """

    def register(cls):
        REGISTERED_COMMANDS[name] = (cls, desc)
        return cls

    return register



"""Recipies to build sample data."""

from . import get_sample_path

RECIPES = {
    "Prep-Type1.h5": lambda path: [
        "heavyedge",
        "prep",
        "--type",
        "csvs",
        "--config",
        get_sample_path("config-prep.yml"),
        get_sample_path("Type1"),
        "-o",
        path,
    ],
    "Prep-Type2.h5": lambda path: [
        "heavyedge",
        "prep",
        "--type",
        "csvs",
        "--config",
        get_sample_path("config-prep.yml"),
        get_sample_path("Type2"),
        "-o",
        path,
    ],
    "Prep-Type3.h5": lambda path: [
        "heavyedge",
        "prep",
        "--type",
        "csvs",
        "--config",
        get_sample_path("config-prep.yml"),
        get_sample_path("Type3"),
        "-o",
        path,
    ],
}



"""Command line interface tools."""

__all__ = [
    "register_command",
    "Command",
    "ConfigArgumentParser",
]

from .command import Command, register_command
from .parser import ConfigArgumentParser



import numpy
from Cython.Build import cythonize
from setuptools import Extension, setup

extensions = [
    Extension(
        "heavyedge._wasserstein",
        ["src/heavyedge/_wasserstein.pyx"],
    ),
]

setup(
    ext_modules=cythonize(extensions),
    include_dirs=[numpy.get_include()],
    include_package_data=True,
)



"""Profile preprocessing."""

import numpy as np
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks
from scipy.stats import linregress

from ..wasserstein import wmean

__all__ = [
    "preprocess",
    "outlier",
    "mean",
]


def preprocess(Y, sigma, std_thres):
    """Preprocess raw profile data.

    Parameters
    ----------
    Y : 1-D array
        1-dimensional profile data.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.
    std_thres : scalar
        Standard deviation threshold to detect contact point.

    Returns
    -------
    Y : 1-D array
        Preprocessed profile data.
    L : int
        Length of *Y* until the contact point.

    Notes
    -----
    Profiles undergo the following steps:

    1. Profile direction is set so that the contact point is on the right hand side.
    2. Contact point is detected, and set to have zero height.

    Examples
    --------
    >>> from heavyedge import get_sample_path, RawProfileCsvs
    >>> from heavyedge.api import preprocess
    >>> Y = next(RawProfileCsvs(get_sample_path("Type3")).profiles())
    >>> Y, L = preprocess(Y, 32, 0.01)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(Y[:L])
    """
    if Y[0] < Y[-1]:
        # Make plateau is on the left and cp is on the right
        Y = np.flip(Y)

    X = np.arange(len(Y))
    h_xx = gaussian_filter1d(Y, sigma, order=2, mode="nearest")
    if len(h_xx) > 0:
        peaks, _ = find_peaks(h_xx)
    else:
        peaks = np.empty(0, dtype=int)

    candidates = []
    for i, x in enumerate(peaks):
        x = X[x:]
        if not len(x) > 2:
            continue
        y = Y[x]
        reg = linregress(x, y)
        residuals = y - (reg.intercept + reg.slope * x)
        std = np.sqrt(np.sum(residuals**2) / (len(x) - 2))
        if std < std_thres:
            candidates.append(i)

    if candidates:
        cp = peaks[candidates[np.argmax(h_xx[peaks[candidates]])]]
    else:
        cp = len(Y) - 1

    Y = Y - Y[cp]
    return Y, cp + 1


def outlier(profiles, thres=3.5):
    """Detect outlier profiles.

    Parameters
    ----------
    profiles : iterable of array
        Profile data, with last point being the contact point.
    thres : scalar, default=3.5
        Z-score threshold for outlier detection.

    Returns
    -------
    is_outlier : array of bool
        Boolean array where True indicates outlier.

    Notes
    -----
    Outliers are detected by applying modified Z-score method [1]_ on cross-sectional
    areas.

    References
    ----------
    .. [1] Boris Iglewicz and David C Hoaglin.
       Volume 16: how to detect and handle outliers. Quality Press, 1993.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import outlier
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as data:
    ...     profiles = list(data.profiles())
    ...     is_outlier = outlier(profiles, 1.5)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... for profile, skip in zip(profiles, is_outlier):
    ...     if skip:
    ...         plt.plot(profile, color="red")
    ...     else:
    ...         plt.plot(profile, alpha=0.2, color="gray")
    """
    x = np.array([np.sum(p) for p in profiles])
    med = np.median(x)
    mad = np.median(np.abs(x - med))
    mod_z = 0.6745 * (x - med) / mad
    return np.abs(mod_z) > thres


def mean(x, profiles, grid_num):
    """Fréchet mean of profiles using Wasserstein distance.

    Parameters
    ----------
    x : ndarray
        X coordinates of *profiles*.
    profiles : list of array
        Profile data, with last point being the contact point.
    grid_num : int
        Number of sample points in [0, 1] to construct regression results.

    Returns
    -------
    ndarray
        Averaged *Y*.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge.api import mean
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as data:
    ...     x = data.x()
    ...     profiles = list(data.profiles())
    >>> mean = mean(x, profiles, 100)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... for profile in profiles:
    ...     plt.plot(profile, alpha=0.2, color="gray")
    ... plt.plot(mean)
    """
    xs, areas, pdfs = [], [], []
    for prof in profiles:
        x_ = x[: len(prof)]
        A = np.trapezoid(prof, x_)
        xs.append(x_)
        areas.append(A)
        pdfs.append(prof / A)
    X, F = wmean(xs, pdfs, grid_num)
    # Fix the last point of X to grid
    last_idx = np.argmin(np.abs(x - X[-1]))
    X[-1] = x[last_idx]

    shape = np.interp(x[: last_idx + 1], X, F)
    shape /= np.trapezoid(shape, x[: len(shape)])
    return shape * np.mean(areas)
